@@ -12,30 +12,42 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+import collections
 import datetime
 
 from oslo_config import cfg
 from oslo_db.sqlalchemy import session as db_session
 from oslo_db.sqlalchemy import utils as db_utils
+import sqlalchemy
+from sqlalchemy.engine.url import make_url
+
+import logging
 
 from subunit2sql.db import models
 from subunit2sql import exceptions
 from subunit2sql import read_subunit
 
 CONF = cfg.CONF
+CONF.register_cli_opt(cfg.BoolOpt('verbose', short='v', default=False,
+                                  help='Verbose output including logging of '
+                                       'SQL statements'))
 
 DAY_SECONDS = 60 * 60 * 24
 
-_FACADE = None
+_facades = {}
 
 
 def _create_facade_lazily():
-    global _FACADE
-    if _FACADE is None:
-        _FACADE = db_session.EngineFacade(
+    global _facades
+    db_url = make_url(CONF.database.connection)
+    db_backend = db_url.get_backend_name()
+    facade = _facades.get(db_backend)
+    if facade is None:
+        facade = db_session.EngineFacade(
             CONF.database.connection,
             **dict(CONF.database.iteritems()))
-    return _FACADE
+        _facades[db_backend] = facade
+    return facade
 
 
 def get_session(autocommit=True, expire_on_commit=False):
@@ -45,8 +57,42 @@ def get_session(autocommit=True, expire_on_commit=False):
     :param bool expire_on_commit: Expire the session on commit defaults False.
     """
     facade = _create_facade_lazily()
-    return facade.get_session(autocommit=autocommit,
-                              expire_on_commit=expire_on_commit)
+    session = facade.get_session(autocommit=autocommit,
+                                 expire_on_commit=expire_on_commit)
+
+    # if --verbose was specified, turn on SQL logging
+    # note that this is done after the session has been initialized so that
+    # we can override the default sqlalchemy logging
+    if CONF.get('verbose', False):
+        logging.basicConfig()
+        logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)
+
+    return session
+
+
+def _filter_runs_by_date(query, start_date=None, stop_date=None):
+    # Helper to apply a data range filter to a query on Run table
+    if isinstance(start_date, str):
+        start_date = datetime.datetime.strptime(start_date, '%Y-%m-%d')
+    if isinstance(stop_date, str):
+        stop_date = datetime.datetime.strptime(stop_date, '%Y-%m-%d')
+    if start_date:
+        query = query.filter(models.Run.run_at >= start_date)
+    if stop_date:
+        query = query.filter(models.Run.run_at <= stop_date)
+    return query
+
+
+def get_engine(use_slave=False):
+    """Get a new sqlalchemy engine instance
+
+    :param bool use_slave if possible, use 'slave' database for this engine
+
+    :return: The engine object for the database connection
+    :rtype: sqlalchemy.engine.Engine
+    """
+    facade = _create_facade_lazily()
+    return facade.get_engine(use_slave=use_slave)
 
 
 def create_test(test_id, run_count=0, success=0, failure=0, run_time=0.0,
@@ -107,7 +153,7 @@ def update_test(values, test_id, session=None):
 
 
 def create_run(skips=0, fails=0, passes=0, run_time=0, artifacts=None,
-               id=None, session=None):
+               id=None, session=None, run_at=None):
     """Create a new run record in the database
 
     :param int skips: total number of skipped tests defaults to 0
@@ -126,6 +172,8 @@ def create_run(skips=0, fails=0, passes=0, run_time=0, artifacts=None,
     run = models.Run()
     if id:
         run.id = id
+    if run_at:
+        run.run_at = run_at
     run.skips = skips
     run.fails = fails
     run.passes = passes
@@ -202,6 +250,7 @@ def add_run_metadata(meta_dict, run_id, session=None):
     :rtype: subunit2sql.models.RunMeta
     """
 
+    session = session or get_session()
     metadata = []
     for key, value in meta_dict.items():
         meta = models.RunMetadata()
@@ -270,10 +319,17 @@ def create_test_run(test_id, run_id, status, start_time=None,
     test_run.test_id = test_id
     test_run.run_id = run_id
     test_run.status = status
-    start_time = start_time.replace(tzinfo=None)
-    start_time_microsecond = start_time.microsecond
-    stop_time = end_time.replace(tzinfo=None)
-    stop_time_microsecond = stop_time.microsecond
+    if start_time:
+        start_time = start_time.replace(tzinfo=None)
+        start_time_microsecond = start_time.microsecond
+    else:
+        start_time_microsecond = None
+    if end_time:
+        stop_time = end_time.replace(tzinfo=None)
+        stop_time_microsecond = stop_time.microsecond
+    else:
+        stop_time = None
+        stop_time_microsecond = None
     test_run.stop_time = stop_time
     test_run.stop_time_microsecond = stop_time_microsecond
     test_run.start_time = start_time
@@ -329,6 +385,51 @@ def get_test_run_metadata(test_run_id, session=None):
     return query.all()
 
 
+def add_test_metadata(meta_dict, test_id, session=None):
+    """Add a metadata key value pairs for a specific test.
+
+    This method will take a dictionary and store key value pair metadata in the
+    DB associated with the specified run.
+
+    :param dict meta_dict: a dictionary which will generate a separate key
+                           value pair row associated with the test_run_id
+    :param str test_id: the uuid of the test to update. (value of the
+                        id column for the row to be updated)
+    :param session: optional session object if one isn't provided a new session
+                    will be acquired for the duration of this operation
+
+    :return list: The list of created metadata objects
+    :rtype: subunit2sql.models.TestMeta
+    """
+    metadata = []
+    for key, value in meta_dict.items():
+        meta = models.TestMetadata()
+        meta.key = key
+        meta.value = value
+        meta.test_id = test_id
+        session = session or get_session()
+        with session.begin():
+            session.add(meta)
+        metadata.append(meta)
+    return metadata
+
+
+def get_test_metadata(test_id, session=None):
+    """Return all test metadata objects for associated with a given test.
+
+    :param str test_id: The uuid of the test to get all the metadata
+    :param session: optional session object if one isn't provided a new session
+                    will be acquired for the duration of this operation
+
+    :return list: The list of created metadata objects
+    :rtype: subunit2sql.models.TestMetadata
+    """
+    session = session or get_session()
+    query = db_utils.model_query(models.TestMetadata, session).filter_by(
+        id=test_id)
+    return query.all()
+
+
 def get_all_tests(session=None):
     """Return all tests from the DB.
 
@@ -343,6 +444,29 @@ def get_all_tests(session=None):
     return query.all()
 
 
+def get_all_runs_by_date(start_date=None, stop_date=None, session=None):
+    """Return all runs from the DB.
+
+    :param str: optional start_date, if provided only runs started at or after
+                the start_date will be included in the response
+    :param str: optional end_date, if provided only runs started at or before
+                the end_date will be included in the response
+    :param session: optional session object if one isn't provided a new session
+                    will be acquired for the duration of this operation
+
+    :return list: The list of run objects
+    :rtype: subunit2sql.models.Run
+    """
+
+    session = session or get_session()
+    query = db_utils.model_query(models.Run, session=session)
+
+    # Process date bounds
+    query = _filter_runs_by_date(query, start_date, stop_date)
+
+    return query.all()
+
+
 def get_all_runs(session=None):
     """Return all runs from the DB.
 
@@ -352,9 +476,8 @@ def get_all_runs(session=None):
     :return list: The list of run objects
     :rtype: subunit2sql.models.Run
     """
-    session = session or get_session()
-    query = db_utils.model_query(models.Run, session)
-    return query.all()
+
+    return get_all_runs_by_date(session=session)
 
 
 def get_all_test_runs(session=None):
@@ -563,23 +686,33 @@ def get_tests_run_dicts_from_run_id(run_id, session=None):
     """
     session = session or get_session()
     query = db_utils.model_query(models.Test, session=session).join(
-        models.TestRun).filter_by(run_id=run_id).join(
-            models.TestRunMetadata).values(
-                models.Test.test_id,
-                models.TestRun.status,
+        models.TestRun).filter(models.TestRun.run_id == run_id).outerjoin(
+            models.TestRunMetadata).order_by(
                 models.TestRun.start_time,
-                models.TestRun.start_time_microsecond,
-                models.TestRun.stop_time,
-                models.TestRun.stop_time_microsecond,
-                models.TestRunMetadata.key,
-                models.TestRunMetadata.value)
-    test_runs = {}
+                models.TestRun.start_time_microsecond).values(
+                    models.Test.test_id,
+                    models.TestRun.status,
+                    models.TestRun.start_time,
+                    models.TestRun.start_time_microsecond,
+                    models.TestRun.stop_time,
+                    models.TestRun.stop_time_microsecond,
+                    models.TestRunMetadata.key,
+                    models.TestRunMetadata.value)
+    test_runs = collections.OrderedDict()
     for test_run in query:
         if test_run[0] not in test_runs:
-            start_time = test_run[2]
-            start_time = start_time.replace(microsecond=test_run[3])
-            stop_time = test_run[4]
-            stop_time = stop_time.replace(microsecond=test_run[5])
+            # If there is no start_time set to None
+            if test_run[2]:
+                start_time = test_run[2]
+                start_time = start_time.replace(microsecond=test_run[3])
+            else:
+                start_time = None
+            # If there is no stop_time set to None
+            if test_run[4]:
+                stop_time = test_run[4]
+                stop_time = stop_time.replace(microsecond=test_run[5])
+            else:
+                stop_time = None
             test_runs[test_run[0]] = {
                 'status': test_run[1],
                 'start_time': start_time,
@@ -635,14 +768,7 @@ def get_test_run_series(start_date=None, stop_date=None, session=None):
         models.RunMetadata).filter_by(key='build_queue', value='gate')
 
     # Process date bounds
-    if isinstance(start_date, str):
-        start_date = datetime.datetime.strptime(start_date, '%b %d %Y')
-    if isinstance(stop_date, str):
-        stop_date = datetime.datetime.strptime(stop_date, '%b %d %Y')
-    if start_date:
-        full_query.filter(models.Run.run_at < start_date)
-    if stop_date:
-        full_query.filter(models.Run.run_at > stop_date)
+    full_query = _filter_runs_by_date(full_query, start_date, stop_date)
 
     query = full_query.values(models.Run.run_at, models.Run.passes,
                               models.Run.fails)
@@ -722,9 +848,14 @@ def delete_old_runs(expire_age=186, session=None):
     """
     session = session or get_session()
     expire_date = datetime.date.today() - datetime.timedelta(days=expire_age)
-    db_utils.model_query(models.Run, session).filter(
-        models.Run.run_at < expire_date).join(
-            models.RunMetadata).delete(synchronize_session='evaluate')
+
+    # Delete the run_metadata
+    sub_query = session.query(models.Run.id).filter(
+        models.Run.run_at < expire_date).subquery()
+    db_utils.model_query(models.RunMetadata, session).filter(
+        models.RunMetadata.run_id.in_(sub_query)).delete(
+            synchronize_session='fetch')
+    # Delete the runs
     db_utils.model_query(models.Run, session).filter(
         models.Run.run_at < expire_date).delete(synchronize_session='evaluate')
 
@@ -738,9 +869,14 @@ def delete_old_test_runs(expire_age=186, session=None):
     """
     session = session or get_session()
     expire_date = datetime.date.today() - datetime.timedelta(days=expire_age)
-    db_utils.model_query(models.TestRun, session).filter(
-        models.TestRun.start_time < expire_date).join(
-            models.TestRunMetadata).delete(synchronize_session='evaluate')
+
+    # Delete the test run metadata
+    sub_query = session.query(models.TestRun.id).filter(
+        models.TestRun.start_time < expire_date).subquery()
+    db_utils.model_query(models.TestRunMetadata, session).filter(
+        models.TestRunMetadata.test_run_id.in_(sub_query)).delete(
+            synchronize_session='fetch')
+    # Delete the test runs
     db_utils.model_query(models.TestRun, session).filter(
         models.TestRun.start_time < expire_date).delete(
             synchronize_session='evaluate')
@@ -820,6 +956,211 @@ def get_test_counts_in_date_range(test_id, start_date=None, stop_date=None,
     return count_dict
 
 
+def get_failing_test_ids_from_runs_by_key_value(key, value, session=None):
+    """Get a list of failing test_ids from runs with run_metadata.
+
+    This method gets a distinct list of test_ids (the test_id column not the id
+    column) from all runs that match a run metadata key value pair.
+
+    :param str key: The key to use to match runs from in run_metadata
+    :param str value: The value of the key in run_metadata to match runs
+                      against
+    :param session: optional session object if one isn't provided a new session
+                    will be acquired for the duration of this operation
+
+    :return: A list of test_ids that failed from runs that match the provided
+             key value run_metadata pair
+    :rtype: list
+    """
+
+    session = session or get_session()
+    test_ids = db_utils.model_query(models.TestRun, session).join(
+        models.Test, models.TestRun.test_id == models.Test.id).join(
+            models.RunMetadata,
+            models.TestRun.run_id == models.RunMetadata.run_id).filter(
+                models.RunMetadata.key == key,
+                models.RunMetadata.value == value,
+                models.TestRun.status == 'fail').values(
+                    sqlalchemy.distinct(models.Test.test_id))
+    return [test_id[0] for test_id in test_ids]
+
+
+def get_test_run_dict_by_run_meta_key_value(key, value, start_date=None,
+                                            stop_date=None, session=None):
+    """Get a list of test run dicts from runs with a run metadata key value pair
+
+    :param str key: The key to use to match runs from in run_metadata
+    :param str value: The value of the key in run_metadata to match runs
+                      against
+    :param start_date: Optional start date to filter results on
+    :param stop_date: Optional stop date to filter results on
+    :param session: Optional session object if one isn't provided a new session
+                    will be acquired for the duration of this operation
+
+    :return test_runs: The dictionary of all the tests run on any run that had
+                       metadata matching the provided key value pair.
+    :rtype: dict
+    """
+    session = session or get_session()
+    query = db_utils.model_query(models.RunMetadata, session).filter(
+        models.RunMetadata.key == key,
+        models.RunMetadata.value == value).join(
+            models.TestRun,
+            models.RunMetadata.run_id == models.TestRun.run_id).join(
+                models.Test)
+    query = _filter_runs_by_date(query, start_date=start_date,
+                                 stop_date=stop_date)
+    query = query.values(models.Test.test_id,
+                         models.TestRun.status,
+                         models.TestRun.start_time,
+                         models.TestRun.start_time_microsecond,
+                         models.TestRun.stop_time,
+                         models.TestRun.stop_time_microsecond)
+    tests = []
+    for test in query:
+        if test[2]:
+            start_time = test[2]
+            start_time = start_time.replace(microsecond=test[3])
+        else:
+            start_time = None
+        if test[4]:
+            stop_time = test[4]
+            stop_time = stop_time.replace(microsecond=test[5])
+        else:
+            stop_time = None
+        test_run_dict = {
+            'test_id': test[0],
+            'status': test[1],
+            'start_time': start_time,
+            'stop_time': stop_time,
+        }
+        tests.append(test_run_dict)
+    return tests
+
+
+def get_all_runs_time_series_by_key(key, start_date=None,
+                                    stop_date=None, session=None):
+    """Get a time series of run summaries grouped by a key
+
+    This method will get a time series dictionary of run summary views which
+    are grouped by the values of the specified key
+
+    :param str key: the key to use for grouping the run summaries
+    :param str start_date: Optional start date to filter results on
+    :param str stop_date: Optional stop date to filter results on
+    :param session: optional session object if one isn't provided a new session
+                    will be acquired for the duration of this operation
+    :return runs: a time series dictionary of runs grouped by values of the
+                  specified key
+    :rtype: dict
+    """
+    session = session or get_session()
+    runs_query = db_utils.model_query(models.Run, session).join(
+        models.RunMetadata).filter(models.RunMetadata.key == key)
+    runs_query = _filter_runs_by_date(runs_query, start_date, stop_date)
+    runs_query = runs_query.values(models.Run.run_at,
+                                   models.Run.passes,
+                                   models.Run.fails,
+                                   models.Run.skips,
+                                   models.RunMetadata.value)
+    runs = {}
+    for run in runs_query:
+        if run[0] not in runs:
+            runs[run[0]] = {run[4]: [{
+                'pass': run[1],
+                'fail': run[2],
+                'skip': run[3],
+            }]}
+        else:
+            if run[4] not in runs[run[0]].keys():
+                runs[run[0]][run[4]] = [{
+                    'pass': run[1],
+                    'fail': run[2],
+                    'skip': run[3],
+                }]
+            else:
+                runs[run[0]][run[4]].append({
+                    'pass': run[1],
+                    'fail': run[2],
+                    'skip': run[3],
+                })
+    return runs
+
+
+def get_time_series_runs_by_key_value(key, value, start_date=None,
+                                      stop_date=None, session=None):
+    """Get a time series of runs with meta for all runs with a key value pai
+
+    :param str key: the metadata key to use for matching the runs
+    :param str value: the metadata value to use for matching the runs
+    :param start_date: Optional start date to filter results on
+    :param str stop_date: Optional stop date to filter results on
+    :param session: optional session object if one isn't provided a new session
+                    will be acquired for the duration of this operation
+
+
+    :return runs: a time series dictionary (where the top level key is a
+                  timestamp) that contains all the runs which
+    :rtype: dict
+    """
+    session = session or get_session()
+    sub_query = session.query(models.RunMetadata.run_id).filter(
+        models.RunMetadata.key == key,
+        models.RunMetadata.value == value).subquery()
+    run_query = db_utils.model_query(models.Run, session).join(
+        models.RunMetadata).filter(models.Run.id.in_(sub_query))
+    run_query = _filter_runs_by_date(run_query, start_date, stop_date)
+    run_query = run_query.values(models.Run.id,
+                                 models.Run.passes,
+                                 models.Run.fails,
+                                 models.Run.skips,
+                                 models.Run.run_time,
+                                 models.Run.run_at,
+                                 models.RunMetadata.key,
+                                 models.RunMetadata.value)
+    runs = {}
+    for run in run_query:
+        run_at = run[5]
+        run_id = run[0]
+        if run_at not in runs:
+            # We have hit a new time stamp so we need to add a top level key
+            # for the timestamp and populate the run list with a new dict for
+            # the run
+            runs[run_at] = []
+            run_dict = {
+                'id': run_id,
+                'pass': run[1],
+                'fail': run[2],
+                'skip': run[3],
+                'run_time': run[4],
+                'metadata': {run[6]: run[7]}
+            }
+            runs[run_at].append(run_dict)
+        else:
+            if run_id not in [loc_run["id"] for loc_run in runs[run_at]]:
+                # We have hit a new run for an existing timestamp, we need to
+                # append a new run dict to the list of runs for that timestamp
+                run_dict = {
+                    'id': run_id,
+                    'pass': run[1],
+                    'fail': run[2],
+                    'skip': run[3],
+                    'run_time': run[4],
+                    'metadata': {run[6]: run[7]}
+                }
+                runs[run_at].append(run_dict)
+            else:
+                # The run dictionary has already been added for this timestamp
+                # this means we've hit a new metadata entry, so we need to
+                # update the metadata dictionary with the additional info
+                update_index = None
+                for index, run_dict in list(enumerate(runs[run_at])):
+                    if run_dict['id'] == run_id:
+                        update_index = index
+                runs[run_at][update_index]['metadata'][run[6]] = run[7]
+    return runs
+
+
 def add_test_run_attachments(attach_dict, test_run_id, session=None):
     """Add attachments a specific test run.
 
@@ -849,3 +1190,77 @@ def add_test_run_attachments(attach_dict, test_run_id, session=None):
             session.add(attachment)
         attachments.append(attachment)
     return attachments
+
+
+def get_runs_by_status_grouped_by_run_metadata(key, start_date=None,
+                                               stop_date=None, session=None):
+    session = session or get_session()
+    val = models.RunMetadata.value
+    run_pass_query = session.query(
+        sqlalchemy.func.count(models.Run.id), val).filter(
+            models.Run.fails == 0, models.Run.passes > 0).join(
+                models.RunMetadata).group_by(val).filter(
+                    models.RunMetadata.key == key)
+    run_fail_query = session.query(
+        sqlalchemy.func.count(models.Run.id), val).filter(
+            models.Run.fails > 0, models.Run.passes > 0).join(
+                models.RunMetadata).group_by(val).filter(
+                    models.RunMetadata.key == key)
+
+    run_pass_query = _filter_runs_by_date(run_pass_query, start_date,
+                                          stop_date)
+    run_fail_query = _filter_runs_by_date(run_fail_query, start_date,
+                                          stop_date)
+    rows = run_pass_query.all()
+    result = {}
+    for row in rows:
+        result[row[1]] = {'pass': row[0]}
+    rows = run_fail_query.all()
+    for row in rows:
+        if row[1] in result:
+            result[row[1]]['fail'] = row[0]
+        else:
+            result[row[1]] = {'fail': row[0]}
+    return result
+
+
+def get_all_run_metadata_keys(session=None):
+    """Get a list of all the keys used in the run_metadata table
+
+    :param session: optional session object if one isn't provided a new session
+                    will be acquired for the duration of this operation
+
+    :return keys: a list of all keys used in the run_metadata table
+    :rtype: list
+    """
+    session = session or get_session()
+    keys = session.query(models.RunMetadata.key).distinct().all()
+    return [key[0] for key in keys]
+
+
+def get_all_test_metadata_keys(session=None):
+    """Get a list of all the keys used in the test_metadata table
+
+    :param session: optional session object if one isn't provided a new session
+                    will be acquired for the duration of this operation
+
+    :return keys: a list of all keys used in the test_metadata table
+    :rtype: list
+    """
+    session = session or get_session()
+    keys = session.query(models.TestMetadata.key).distinct().all()
+    return [key[0] for key in keys]
+
+
+def get_all_test_run_metadata_keys(session=None):
+    """Get a list of all the keys used in the test_run_metadata table
+
+    :param session: optional session object if one isn't provided a new session
+                    will be acquired for the duration of this operation
+
+    :return keys: a list of all keys used in the test_run_metadata table
+    :rtype: list
+    """
+    session = session or get_session()
+    keys = session.query(models.TestRunMetadata.key).distinct().all()
+    return [key[0] for key in keys]
