@@ -17,65 +17,17 @@
 import ConfigParser
 import datetime
 import os
-import subprocess
 
 
 from alembic import config
 from alembic import script
-from oslo_config import cfg
-from oslo_db import options
-import six
-from six.moves.urllib import parse
 import sqlalchemy
+from sqlalchemy.engine import reflection
 
 from subunit2sql import exceptions as exc
-from subunit2sql.migrations import cli
 from subunit2sql.tests import base
-
-CONF = cfg.CONF
-CONF.register_cli_opts(options.database_opts, group='database')
-
-
-def _get_connect_string(backend,
-                        user="openstack_citest",
-                        passwd="openstack_citest",
-                        database="openstack_citest"):
-    """Generate a db uri for testing locally.
-
-    Try to get a connection with a very specific set of values, if we get
-    these then we'll run the tests, otherwise they are skipped
-    """
-    if backend == "mysql":
-        backend = "mysql+mysqldb"
-    elif backend == "postgres":
-        backend = "postgresql+psycopg2"
-
-    return ("%(backend)s://%(user)s:%(passwd)s@localhost/%(database)s"
-            % {'backend': backend, 'user': user, 'passwd': passwd,
-               'database': database})
-
-
-def _is_backend_avail(backend,
-                      user="openstack_citest",
-                      passwd="openstack_citest",
-                      database="openstack_citest"):
-    try:
-        if backend == "mysql":
-            connect_uri = _get_connect_string("mysql", user=user,
-                                              passwd=passwd, database=database)
-        elif backend == "postgres":
-            connect_uri = _get_connect_string("postgres", user=user,
-                                              passwd=passwd, database=database)
-        engine = sqlalchemy.create_engine(connect_uri)
-        connection = engine.connect()
-    except Exception:
-        # intentionally catch all to handle exceptions even if we don't
-        # have any backend code loaded.
-        return False
-    else:
-        connection.close()
-        engine.dispose()
-        return True
+from subunit2sql.tests import db_test_utils
+from subunit2sql.tests import subunit2sql_fixtures as fixtures
 
 
 def get_table(engine, name):
@@ -123,173 +75,44 @@ class TestWalkMigrations(base.TestCase):
         for key, value in self.test_databases.items():
             self.engines[key] = sqlalchemy.create_engine(value)
 
-        # We start each test case with a completely blank slate.
-        self._reset_databases()
-
     def assertColumnExists(self, engine, table, column):
         table = get_table(engine, table)
         self.assertIn(column, table.c)
 
-    def _reset_databases(self):
-        def execute_cmd(cmd=None):
-            proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
-                                    stderr=subprocess.STDOUT, shell=True)
-            output = proc.communicate()[0]
-            self.assertEqual(0, proc.returncode, 'Command failed with '
-                             'output:\n%s' % output)
-        for key, engine in self.engines.items():
-            conn_string = self.test_databases[key]
-            conn_pieces = parse.urlparse(conn_string)
-            engine.dispose()
-            if conn_string.startswith('sqlite'):
-                # We can just delete the SQLite database, which is
-                # the easiest and cleanest solution
-                db_path = conn_pieces.path[1:]
-                if os.path.exists(db_path):
-                    os.unlink(db_path)
-                # No need to recreate the SQLite DB. SQLite will
-                # create it for us if it's not there...
-            elif conn_string.startswith('mysql'):
-                # We can execute the MySQL client to destroy and re-create
-                # the MYSQL database, which is easier and less error-prone
-                # than using SQLAlchemy to do this via MetaData...trust me.
-                database = conn_pieces.path.strip('/')
-                loc_pieces = conn_pieces.netloc.split('@')
-                host = loc_pieces[1]
-                auth_pieces = loc_pieces[0].split(':')
-                user = auth_pieces[0]
-                password = ""
-                if len(auth_pieces) > 1:
-                    if auth_pieces[1].strip():
-                        password = "-p\"%s\"" % auth_pieces[1]
-                sql = ("drop database if exists %(database)s; create "
-                       "database %(database)s;") % {'database': database}
-                cmd = ("mysql -u \"%(user)s\" %(password)s -h %(host)s "
-                       "-e \"%(sql)s\"") % {'user': user, 'password': password,
-                                            'host': host, 'sql': sql}
-                execute_cmd(cmd)
-            elif conn_string.startswith('postgresql'):
-                database = conn_pieces.path.strip('/')
-                loc_pieces = conn_pieces.netloc.split('@')
-                host = loc_pieces[1]
-
-                auth_pieces = loc_pieces[0].split(':')
-                user = auth_pieces[0]
-                password = ""
-                if len(auth_pieces) > 1:
-                    password = auth_pieces[1].strip()
-                # note(boris-42): This file is used for authentication
-                # without password prompt.
-                createpgpass = ("echo '*:*:*:%(user)s:%(password)s' > "
-                                "~/.pgpass && chmod 0600 ~/.pgpass" %
-                                {'user': user, 'password': password})
-                execute_cmd(createpgpass)
-                # note(boris-42): We must create and drop database, we can't
-                # drop database which we have connected to, so for such
-                # operations there is a special database template1.
-                sqlcmd = ("psql -w -U %(user)s -h %(host)s -c"
-                          " '%(sql)s' -d template1")
-                sql = ("drop database if exists %(database)s;")
-                sql = sql % {'database': database}
-                droptable = sqlcmd % {'user': user, 'host': host,
-                                      'sql': sql}
-                execute_cmd(droptable)
-                sql = ("create database %(database)s;")
-                sql = sql % {'database': database}
-                createtable = sqlcmd % {'user': user, 'host': host,
-                                        'sql': sql}
-                execute_cmd(createtable)
-
-    def _get_alembic_config(self, uri):
-        db_config = config.Config(os.path.join(self.script_location,
-                                               'alembic.ini'))
-        db_config.set_main_option('script_location', 'subunit2sql:migrations')
-        db_config.subunit2sql_config = CONF
-        db_config.subunit2sql_config.set_override('connection',
-                                                  six.text_type(uri),
-                                                  group='database')
-        self.script_dir = script.ScriptDirectory.from_config(db_config)
-        return db_config
-
-    def _revisions(self, downgrade=False):
+    def _revisions(self):
         """Provides revisions and its parent revisions.
 
-        :param downgrade: whether to include downgrade behavior or not.
-        :type downgrade: Bool
         :return: List of tuples. Every tuple contains revision and its parent
         revision.
         """
-
-        revisions = list(self.script_dir.walk_revisions("base", "head"))
-
-        if not downgrade:
-            revisions = list(reversed(revisions))
+        db_config = config.Config(os.path.join(self.script_location,
+                                               'alembic.ini'))
+        db_config.set_main_option('script_location', 'subunit2sql:migrations')
+        script_dir = script.ScriptDirectory.from_config(db_config)
+        revisions = list(script_dir.walk_revisions("base", "head"))
 
         if not revisions:
             raise exc.DbMigrationError('There is no suitable migrations.')
 
-        for rev in revisions:
-            if downgrade:
-                # Destination, current
-                yield rev.down_revision, rev.revision
-            else:
-                # Destination, current
-                yield rev.revision, rev.down_revision
+        for rev in list(reversed(revisions)):
+            # Destination, current
+            yield rev.revision, rev.down_revision
 
-    def _walk_versions(self, config, engine, downgrade=True, snake_walk=False):
-        """Test migrations ability to upgrade and downgrade.
-
-        :param downgrade: whether to include downgrade behavior or not.
-        :type downgrade: Bool
-        :snake_walk: enable mode when at every upgrade revision will be
-        downgraded and upgraded in previous state at upgrade and backward at
-        downgrade.
-        :type snake_walk: Bool
-        """
+    def _walk_versions(self, engine):
+        """Test migrations ability to upgrade."""
 
         revisions = self._revisions()
         for dest, curr in revisions:
-            self._migrate_up(config, engine, dest, curr, with_data=True)
+            self._migrate_up(engine, dest, curr, with_data=True)
 
-            if snake_walk and dest != 'None':
-                # NOTE(I159): Pass reversed arguments into `_migrate_down`
-                # method because we have been upgraded to a destination
-                # revision and now we going to downgrade back.
-                self._migrate_down(config, curr, dest, with_data=True)
-                self._migrate_up(config, dest, curr, with_data=True)
-
-        if downgrade:
-            revisions = self._revisions(downgrade)
-            for dest, curr in revisions:
-                self._migrate_down(config, engine, dest, curr, with_data=True)
-                if snake_walk:
-                    self._migrate_up(config, engine, curr, dest,
-                                     with_data=True)
-                    self._migrate_down(config, engine, dest, curr,
-                                       with_data=True)
-
-    def _migrate_down(self, config, engine, dest, curr, with_data=False):
-
-        if dest:
-            cli.do_alembic_command(config, 'downgrade', dest)
-        else:
-            meta = sqlalchemy.MetaData(bind=engine)
-            meta.drop_all()
-
-        if with_data:
-            post_downgrade = getattr(
-                self, "_post_downgrade_%s" % curr, None)
-            if post_downgrade:
-                post_downgrade(engine)
-
-    def _migrate_up(self, config, engine, dest, curr, with_data=False):
+    def _migrate_up(self, engine, dest, curr, with_data=False):
         if with_data:
             data = None
             pre_upgrade = getattr(
                 self, "_pre_upgrade_%s" % dest, None)
             if pre_upgrade:
                 data = pre_upgrade(engine)
-        cli.do_alembic_command(config, 'upgrade', dest)
+        db_test_utils.run_migration(dest, engine)
         if with_data:
             check = getattr(self, "_check_%s" % dest, None)
             if check and data:
@@ -302,8 +125,7 @@ class TestWalkMigrations(base.TestCase):
         that there are no errors in the version scripts for each engine
         """
         for key, engine in self.engines.items():
-            config = self._get_alembic_config(self.test_databases[key])
-            self._walk_versions(config, engine, self.snake_walk)
+            self._walk_versions(engine)
 
     def test_mysql_connect_fail(self):
         """Test graceful mysql connection failure.
@@ -311,24 +133,25 @@ class TestWalkMigrations(base.TestCase):
         Test that we can trigger a mysql connection failure and we fail
         gracefully to ensure we don't break people without mysql
         """
-        if _is_backend_avail('mysql', user="openstack_cifail"):
+        if db_test_utils.is_backend_avail('mysql', user="openstack_cifail"):
             self.fail("Shouldn't have connected")
 
     def test_mysql_opportunistically(self):
+        if not db_test_utils.is_backend_avail('mysql'):
+            raise self.skipTest('mysql is not available')
+
+        self.useFixture(fixtures.LockFixture('mysql'))
+        self.useFixture(fixtures.MySQLConfFixture())
         # Test that table creation on mysql only builds InnoDB tables
-        if not _is_backend_avail('mysql'):
-            self.skipTest("mysql not available")
         # add this to the global lists to make reset work with it, it's removed
         # automatically in tearDown so no need to clean it up here.
-        connect_string = _get_connect_string("mysql")
+        connect_string = db_test_utils.get_connect_string("mysql")
         engine = sqlalchemy.create_engine(connect_string)
-        config = self._get_alembic_config(connect_string)
         self.engines["mysqlcitest"] = engine
         self.test_databases["mysqlcitest"] = connect_string
 
         # build a fully populated mysql database with all the tables
-        self._reset_databases()
-        self._walk_versions(config, engine, False, False)
+        self._walk_versions(engine)
 
         connection = engine.connect()
         # sanity check
@@ -352,24 +175,36 @@ class TestWalkMigrations(base.TestCase):
         Test that we can trigger a postgres connection failure and we fail
         gracefully to ensure we don't break people without postgres
         """
-        if _is_backend_avail('postgresql', user="openstack_cifail"):
+        if db_test_utils.is_backend_avail('postgresql',
+                                          user="openstack_cifail"):
             self.fail("Shouldn't have connected")
 
     def test_postgresql_opportunistically(self):
         # Test postgresql database migration walk
-        if not _is_backend_avail('postgres'):
-            self.skipTest("postgresql not available")
+        if not db_test_utils.is_backend_avail('postgres'):
+            raise self.skipTest('postgres is not available')
+        self.useFixture(fixtures.LockFixture('postgres'))
+        self.useFixture(fixtures.PostgresConfFixture())
         # add this to the global lists to make reset work with it, it's removed
         # automatically in tearDown so no need to clean it up here.
-        connect_string = _get_connect_string("postgres")
+        connect_string = db_test_utils.get_connect_string("postgres")
         engine = sqlalchemy.create_engine(connect_string)
-        config = self._get_alembic_config(connect_string)
         self.engines["postgresqlcitest"] = engine
         self.test_databases["postgresqlcitest"] = connect_string
 
         # build a fully populated postgresql database with all the tables
-        self._reset_databases()
-        self._walk_versions(config, engine, False, False)
+        self._walk_versions(engine)
+
+    def test_sqlite_opportunistically(self):
+        self.useFixture(fixtures.LockFixture('sqlite'))
+        self.useFixture(fixtures.SqliteConfFixture())
+
+        connect_string = db_test_utils.get_connect_string("sqlite")
+        engine = sqlalchemy.create_engine(connect_string)
+        self.engines["sqlitecitest"] = engine
+        self.test_databases["sqlitecitest"] = connect_string
+
+        self._walk_versions(engine)
 
     def _pre_upgrade_1f92cfe8a6d3(self, engine):
         tests = get_table(engine, 'tests')
@@ -525,3 +360,41 @@ class TestWalkMigrations(base.TestCase):
         else:
             self.assertEqual(start_micro, row.start_time_microsecond)
             self.assertEqual(row.stop_time_microsecond, stop_micro)
+
+    def _pre_upgrade_2fb76f1a1393(self, engine):
+        test_metadata = get_table(engine, 'test_metadata')
+        tests = get_table(engine, 'tests')
+        test = {'id': 'fake_test_with_metadata.id',
+                'test_id': 'fake_project.tests.FakeTestClass.fake_test_meta',
+                'run_count': 1,
+                'success': 1,
+                'failure': 0}
+        tests.insert().values(test).execute()
+        data = {'id': 'AUUID',
+                'key': 'AKey',
+                'value': 42,
+                'test_run_id': 'fake_test_with_metadata.id'}
+        test_metadata.insert().values(data).execute()
+        return data
+
+    def _check_2fb76f1a1393(self, engine, data):
+        test_metadata = get_table(engine, 'test_metadata')
+        res = list(test_metadata.select().execute())[0]
+        self.assertEqual(res.id, data['id'])
+        self.assertEqual(res.test_id, data['test_run_id'])
+
+    def _pre_upgrade_b96122f780(self, engine):
+        # NOTE(mtreinish) Return fake data to ensure we run check, this
+        # is needed because the framework normall assumes you're preseeded
+        # data is in the correct state post migration. But this time all we
+        # want to do is ensure a single index exists so that isn't needed
+        return 'data'
+
+    def _check_b96122f780(self, engine, data):
+        insp = reflection.Inspector(engine)
+        indxs = insp.get_indexes('test_runs')
+        # Check that we don't duplicate indexes anymore
+        tests = [indx for indx in indxs if ['test_id'] == indx['column_names']]
+        runs = [indx for indx in indxs if indx['column_names'] == ['run_id']]
+        self.assertEqual(1, len(tests))
+        self.assertEqual(1, len(runs))
